/*******************************************************************************
 * Copyright (C) 2018 THL A29 Limited, a Tencent company.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Tencent - support multiple contexts
 ******************************************************************************/
package com.eclipsesource.v8;

import com.eclipsesource.v8.utils.MemoryManager;

/**
 * Created by LiYongjie on 2018/3/13.
 */
public final class MultiContextV8 {
    private final V8 v8;
    private final V8ContextSessionMgr contextSessionMgr;

    private MultiContextV8() {
        this(null);
    }

    private MultiContextV8(final String globalAlias) {
        v8 = V8.createV8Runtime(globalAlias, null, false);
        contextSessionMgr = new V8ContextSessionMgr(getV8());
    }

    public static MultiContextV8 createMultiContextV8() {
        return new MultiContextV8();
    }

    public static MultiContextV8 createMultiContextV8(final String globalAlias) {
        return new MultiContextV8(globalAlias);
    }

    public V8Context createContext(int contextTag) {
        long v8ContextPtr = v8.createV8Context(contextTag);
        return new V8ContextWrapper(this, v8ContextPtr).context();
    }

    public void release() {
        v8.release();
    }

    public long getIsolatePtr() {
        return v8.getIsolatePtr();
    }

    public V8Locker getV8Locker() {
        return v8.getLocker();
    }

    //TODO
    public MemoryManager createMemoryManager() {
        return new MemoryManager(v8);
    }

    protected void enterContext(V8ContextWrapper context) {
        contextSessionMgr.enterContext(context);
    }

    protected void releaseContext(V8ContextWrapper context) {
        contextSessionMgr.releaseContext(context);
    }

    protected V8 getV8() {
        return v8;
    }
}
