#基于J2V8的多context改造方案

##背景
1. 基于安全或者性能的考虑，多方需要使用v8的多context特性，包括小程序，小游戏，小程序widget，小程序nodejs等。
2. 微信当前存在多套操作v8的接口方案，经过对比，想统一成J2V8的方案。J2V8是一个开源的库，提供了丰富且号称高性能的Java操作v8的接口，使得可以使用java代码实现js代码的操作。
3. **J2V8只是支持单context，这里需要进行多context改造**

##V8::Context简介
V8的context提供了一个隔离的js执行环境，一个V8可以有多个context。context的切换如下图：

![Context Switch](https://mmbiz.qpic.cn/mmbiz_png/4wQxza4CwrRFc59ZJSIjicFHepDa3bLGGpUz5JYiaqIc5npJDesuYwfm9SNKNmiaiaYdjkVoFmG79QvQ6UP4buCcOw/640?wx_fmt=png)

##J2V8多context改造简介
###J2V8的单context方案
J2V8提供了丰富的java接口，最终都归一到V8.java中一系列jni接口。而**这些jni接口的实现，基本会先调一个SETUP的宏**。这里以executeVoidScript为例子分析一下：

```
JNIEXPORT void JNICALL Java_com_eclipsesource_v8_V8__1executeVoidScript
(JNIEnv * env, jobject v8, jlong v8RuntimePtr, jstring jjstring, jstring jscriptName = NULL, jint jlineNumber = 0) {
  Isolate* isolate = SETUP(env, v8RuntimePtr, );
  ...
```
再看SETUP宏：

```
#define SETUP(env, v8RuntimePtr, errorReturnResult) getIsolate(env, v8RuntimePtr);\
    ...
    HandleScope handle_scope(isolate);\
    Local<Context> context = Local<Context>::New(isolate,runtime->context_);\
    Context::Scope context_scope(context);
```

结合前面V8::context，**SETUP宏的作用之一是，每次接口调用，都先进入runtime->context_，执行完逻辑后退出这个context**。

runtime->context_对应的结构体为：

```
class V8Runtime {
public:
  Isolate* isolate;
  Persistent<Context> context_;
  Persistent<Object>* globalObject;
  ...
}
```
再找到它的初始化地方：

```
JNIEXPORT jlong JNICALL Java_com_eclipsesource_v8_V8__1createIsolate
 (JNIEnv *env, jobject v8, jstring globalAlias, jboolean createWithContext) {
  ...
      Handle<Context> context = Context::New(runtime->isolate, NULL, globalObject);
      runtime->context_.Reset(runtime->isolate, context);
  ...
  return reinterpret_cast<jlong>(runtime);
}
```
createIsolate 对应JAVA接口 V8.createV8Runtime()。
这也就说V8的一个runtime对应了一个context，后面接口都在这个context下执行。

###基于J2V8的多context实现
原生的J2V8支持单context，为支持多context，需要进行改造。  
改造方案基于以下几点考虑：   
1. 改动成本尽可能小  
2. 多context状态不容易混乱  
实现方案如下：

![MultiContext Diagram](https://mmbiz.qpic.cn/mmbiz_png/4wQxza4CwrRFc59ZJSIjicFHepDa3bLGGbtzcmab6Hltib3bqgclXtfiaM1KiaibTibwrMy0yK3RYcHNvHqS2ictk7fJg/640?wx_fmt=png)

简单说明一下：  
java层主要提供了MultiContextV8和V8Context给开发者使用。V8Context提供了一系列java接口。  
1. 开发者调用V8Context的接口时，利用类似AOP的技术，会统一先调用到ContextSessionMgr的enterContext，对应jni的switchContext，这时只是设置runtime-context_为当前对应的context对象，并没有进入对应的context。  
2. 切面执行后，继续执行实际的接口，并调用对应的JNI接口。这时会先调用SETUP宏，如前面分析，这时会进入runtime-context\_。  
3. 由1、2，从V8Context调用接口，保证了在当前context执行。  

好处：  
1. 这个方案不需要每次执行接口都传递contextd到jni层 -> 从而改动成本小。  
2. 也不需要开发者去维护context的进入和退出状态 -> 保证多context状态安全。  

###V8Context接口简介
MultiContextV8主要包括：  
1. 基础：createMultiContextV8、createContext、release  

V8Context 主要包括以下几类接口：  
1. context释放操作：release、isReleased  
2. js执行：executeVoidScript、executeIntegerScript等等  
3. js函数执行：executeIntegerFunction、executeStringFunction等等  
4. js object getter：getInteger、getObject等等（receiver为全局对象）  
5. js object adder：add、addUndefined等等（receiver为全局对象）  
6. 注册js上回调：registerJavaMethod（receiver为全局对象）  
7. context间通信：shareObject 
8. new 对象：newV8Object、newV8Array等等 
##多Context安全
J2V8保证了一个V8 runtime在一个线程执行，所有没有多线程的安全问题。这里讨论一下多context安全。  
首先给结论：改造后的库，**多context的需求使用MultiContextV8类和V8Context类，单context的需求依然可以使用原生的V8类**，并且所有可以混乱context的状态接口并没有暴露出来，所以可以随便使用，context状态也不会混乱，包括其他类比如V8Object，V8Array等。  
看个例子：

```
MultiContextV8 v8 = MultiContextV8.createMultiContextV8();
V8Context context1 = v8.createContext();
V8Context context2 = v8.createContext();
context1.executeVoidScript("var share={a:33};"); // 有切面，runtime->context_ set to context1
V8Object v8Object = context1.getObject("share"); // 有切面，不需要切换
context2.executeVoidScript("var share={a:77};"); // 有切面，runtime->context_ set to context2
int a = v8Object.getInteger("a"); // v8Object没有切面，没切换，不过依然是context1的值，a=33
```

不具体分析了，就是不用刻意去理解context的状态切换，使用V8Object和V8Array等等其他J2V8提供的类，也不会有context状态切换带来的问题。

## 总结
* 接口设计及稳定性、性能等还依赖具体落地时逐步优化

``